#include "allegro.h"
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include "version.h"
#include "prototypes.h"

int gDebugMode=0,gFGColor,gBGColor;
BITMAP *gBackImage;
PALLETE gBackPalette;
char gBackName[256];
int gForceSpeed=0;

extern int pixdbl;
extern int optind;
extern char *optarg;
extern int getopt(int argc, char * const * argv, const char *optstring);
 
int def_fg,def_bg,img_fg,img_bg;

int prepare(char *msg,char *param)
{      
   printf(msg,param);
   printf("Press any to key to start\n");
   readkey();
   set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
   /* select the bitmap pallete */
   if(gBackImage)
   {
      set_pallete(gBackPalette);
      gFGColor=img_fg;
      gBGColor=img_bg;
   }
   else
   {         
      gFGColor=def_fg;
      gBGColor=def_bg;
      set_pallete(desktop_pallete);
   }
}

int main(int argc, char *argv[])
{
  int c;

  strcpy(gBackName,"vmu.pcx");
  def_fg=255;
  def_bg=0;
  img_fg=13;
  img_bg=132;

   while((c=getopt(argc, argv, ":dhV123s:I:F:B:"))!=EOF)
   {
      switch(c)
      {
      case 'h':
         fprintf(stderr, "Usage: %s [-h][-d][-V][-1][-2][-3][-s:nnn][-I:fff][-B:nnn][-F:nnn] game.vms\n"
                         "       -h     Displays this info\n"
                         "       -d     Debug mode\n"
                         "       -V     Display version number\n"
                         "       -1     Single pixel display mode\n"
                         "       -2     Double pixel display mode\n"
                         "       -3     Triple pixel display mode (default)\n"
                         "       -s:NNN Change throttle speed to NNN (numeric)\n"
                         "       -I:FFF Change background image to FFF (must be .PCX file)\n"
                         "       -F:NNN Change colour index used as foreground colour to NNN\n"
                         "       -B:NNN Change colour index used as background colour to NNN\n",argv[0]);
         break;
      case 's':
         gForceSpeed=atoi(optarg+1);
         fprintf(stderr, "Speed forced to %d\n",gForceSpeed);
         break;
      case 'I':
         strcpy(gBackName,optarg+1);
         fprintf(stderr, "Background image changed to %s\n",gBackName);
         break;
      case 'F':
         def_fg=atoi(optarg+1);
         img_fg=atoi(optarg+1);
         fprintf(stderr, "Foreground color changed to %d\n",def_fg);
         break;
      case 'B':
         def_bg=atoi(optarg+1);
         img_bg=atoi(optarg+1);
         fprintf(stderr, "Background color changed to %d\n",def_bg);
         break;
      case 'V':
         fprintf(stderr, "softvms " VERSION " by Marcus Comstedt <marcus@idonex.se>\n");
         break;
      case '1':
         pixdbl = 0;
         break;
      case '2':
         pixdbl = 1;
         break;
      case '3':
         pixdbl = 2;
         break;
      case 'd':
         gDebugMode=1;
         break;
      }
   }
   gBackImage = load_bitmap(gBackName,gBackPalette);
   if (!gBackImage)
      printf("Error reading bitmap file '%s'\n", gBackName);

   allegro_init();
   install_keyboard(); 
   text_mode(0);

   if(argc>optind)
   {
      int fd = open(argv[optind], O_RDONLY|O_BINARY);
      int r;
      char hdr[4];
      if(fd<0 || (r=read(fd, hdr, 4)<4))
      {
         perror(argv[optind]);
         if(fd>=0)
            close(fd);
         clear_keybuf();
         return 1;
      }

      if(hdr[0]=='L' && hdr[1]=='C' && hdr[2]=='D' && hdr[3]=='i')
      {     
         prepare("Starting with lcdimage file:%s\n",argv[optind]);
         do_lcdimg(argv[optind]);
      }
      else
      {      
         prepare("Starting with vmsgame:%s\n",argv[optind]);
         do_vmsgame(argv[optind]);
      }
   }
   else
   {      
      prepare("No game/lcdimage specified.\n",NULL);
      clear(screen);
      /* blit the image onto the screen */
      if(gBackImage)
         blit(gBackImage,screen,0,0,(SCREEN_W-gBackImage->w)/2,0,gBackImage->w,gBackImage->h);
	   text_mode(-1);
      textprintf_centre(screen, font,SCREEN_W/2, 120, gFGColor, "NO GAME LOADED");
      textprintf_centre(screen, font,SCREEN_W/2, 150, gFGColor, "HIT ESC TO EXIT");
		for(;;)
      {
         waitforevents(NULL);
         checkevents();
      }
   }
   clear_keybuf();
   return 0;
}


